'''
Miris Manager client library
This module is not intended to be used directly, only the client class should be used.
'''
import json
import logging
import re
from pathlib import Path

from ..conf import BASE_CONF

logger = logging.getLogger('mm_client.lib.configuration')


def load_conf(default_conf=None, local_conf=None):
    # copy default configuration
    conf = BASE_CONF.copy()
    # update with default and local configuration
    for index, conf_override in enumerate((default_conf, local_conf)):
        if not conf_override:
            continue
        if isinstance(conf_override, str):
            conf_override = Path(conf_override)
        if isinstance(conf_override, dict):
            for key, val in conf_override.items():
                if not key.startswith('_'):
                    conf[key] = val
        elif isinstance(conf_override, Path):
            if conf_override.exists():
                content = conf_override.read_text()
                content = re.sub(r'\n\s*//.*', '\n', content)  # remove comments
                conf_mod = json.loads(content) if content else None
                if not conf_mod:
                    logger.debug('Config file "%s" is empty.', conf_override)
                else:
                    logger.debug('Config file "%s" loaded.', conf_override)
                    if not isinstance(conf_mod, dict):
                        raise ValueError(f'The configuration in "{conf_override}" is not a dict.')
                    conf.update(conf_mod)
            else:
                logger.debug('Config file does not exists, using default config.')
        else:
            raise ValueError('Unsupported type for configuration.')
    if conf['SERVER_URL'].endswith('/'):
        conf['SERVER_URL'] = conf['SERVER_URL'].rstrip('/')
    return conf


def update_conf(local_conf, key, value):
    if not local_conf:
        logger.debug('Cannot update configuration, "local_conf" is not set.')
        return False
    if isinstance(local_conf, str):
        local_conf = Path(local_conf)
    elif not isinstance(local_conf, Path):
        logger.debug('Cannot update configuration, "local_conf" is not a path.')
        return False

    if local_conf.is_file():
        content = local_conf.read_text().strip()
    else:
        content = ''
    data = json.loads(content) if content else {}
    data[key] = value
    new_content = json.dumps(data, sort_keys=True, indent=4)
    local_conf.write_text(new_content)
    logger.debug('Configuration file "%s" updated: "%s" set to "%s".', local_conf, key, value)
    return True


def check_conf(conf):
    # check that mandatory configuration values are set
    if not conf.get('SERVER_URL') or conf['SERVER_URL'] == 'https://mirismanager':
        raise ValueError('The value of "SERVER_URL" is not set. Please configure it.')
    conf['SERVER_URL'] = conf['SERVER_URL'].strip('/')



'''
Miris Manager client library
This module is not intended to be used directly, only the client class should be used.
'''
import logging
import socket
import uuid
import os

logger = logging.getLogger('mm_client.lib.info')


def get_host_info(url):
    '''
    Collect information on local system.
    '''
    # get hostname
    hostname = socket.gethostname()
    # get local IP address
    host = url.split('://')[-1]
    if ':' in host:
        host, port = host.split(':')
        port = int(port)
    elif url.startswith('http:'):
        port = 80
    else:
        port = 443
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    if host.endswith('/'):
        host = host[:-1]
    try:
        s.connect((host, port))
    except Exception as e:
        s.close()
        raise e
    local_ip = s.getsockname()[0]
    s.close()
    # get MAC address
    node = uuid.getnode()
    mac = ':'.join(('%012x' % node)[i:i + 2] for i in range(0, 12, 2))
    info = dict(
        hostname=hostname,
        local_ip=local_ip,
        mac=mac,
    )
    logger.debug('[%s] Client info is %s' % (host, info))
    return info


def get_free_space_bytes(path):
    '''
    Get free space on partition used for given path.
    '''
    statvfs = os.statvfs(path)
    free = statvfs.f_frsize * statvfs.f_bavail
    return free


def get_remaining_space():
    '''
    Return remaining space in /home in MB.
    '''
    return int(get_free_space_bytes('/home') / 1000000)



import datetime

import pytest


def test_signature__default():
    from mm_client.lib.signing import get_signature, check_signature

    conf = {}

    signature = get_signature(conf)
    assert signature == {}

    assert check_signature(conf, {}) is None


def test_signature__configured():
    from mm_client.lib.signing import get_signature, check_signature

    conf = {
        'SECRET_KEY': 'the secret key',
        'API_KEY': 'the API key',
    }

    signature = get_signature(conf)
    assert sorted(signature.keys()) == ['hmac', 'time']

    assert check_signature(conf, signature) is None


@pytest.mark.parametrize('signature, expected', [
    pytest.param(
        {},
        'some mandatory data are missing.',
        id='missing'),
    pytest.param(
        {'time': 'invalid', 'hmac': 'test'},
        'the received time is invalid.',
        id='date'),
    pytest.param(
        {'time': '2000-01-01_00-00-00_000', 'hmac': 'test'},
        'the difference between the request time and the current time is too large.',
        id='expired'),
    pytest.param(
        {'time': datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S_%f'), 'hmac': 'test'},
        'the received and computed HMAC values do not match.',
        id='hmac'),
])
def test_signature__invalid(signature, expected):
    from mm_client.lib.signing import check_signature

    conf = {
        'SECRET_KEY': 'the secret key',
        'API_KEY': 'the API key',
    }

    assert check_signature(conf, signature) == expected



# Miris Manager client base configuration
# This file should not be modified directly, put your modification in another json file and give the path to the client.

BASE_CONF = {
    # Logging level
    'LOG_LEVEL': 'INFO',

    # Server URL of Miris Manager
    'SERVER_URL': 'https://mirismanager',

    # API key of this system in Miris Manager
    # The API key is automatically set when empty and when Capus Manager discovery mode is enabled.
    'API_KEY': '',

    # Secret key of this system in Miris Manager, used to sign messages
    'SECRET_KEY': '',

    # Try to register this system if no API_KEY is defined
    'AUTO_REGISTRATION': True,

    # Notify systemd watchdog after each long polling call
    'WATCHDOG': False,

    # Verify server SSL certificate
    'VERIFY_SSL': False,

    # API requests max duration in seconds
    'TIMEOUT': 10,

    # Proxies for API requests
    # To use system proxies: None (proxies should be set in environment)
    # To disable proxies: {'http': '', 'https': ''}
    # To use a proxy: {'http': 'http://10.10.1.10:3128', 'https': 'http://10.10.1.10:1080'}
    'PROXIES': None,

    # This list makes available or not actions buttons in Miris Manager
    'CAPABILITIES': {},

    # List of Miris Manager urls (do not overwritte this)
    'API_CALLS': {
        'PING': {'method': 'get', 'url': '/api/', 'anonymous': True},
        'TIME': {'method': 'get', 'url': '/api/time/', 'anonymous': True},
        'INFO': {'method': 'get', 'url': '/api/info/', 'anonymous': True},
        'LONG_POLLING': {'method': 'get', 'url': '/remote-event/v3'},
        'SET_COMMAND_STATUS': {'method': 'post', 'url': '/api/v3/fleet/control/set-command-status/'},
        'GET_INFO': {'method': 'get', 'url': '/api/v3/fleet/systems/get-info/'},
        'SET_INFO': {'method': 'post', 'url': '/api/v3/fleet/systems/set-info/'},
        'GET_STATUS': {'method': 'get', 'url': '/api/v3/fleet/systems/get-status/'},
        'SET_STATUS': {'method': 'post', 'url': '/api/v3/fleet/systems/set-status/'},
        'SET_SCREENSHOT': {'method': 'post', 'url': '/api/v3/fleet/systems/set-screenshot/'},
        'REGISTER_SYSTEM': {'method': 'post', 'url': '/api/v3/fleet/systems/register/'},
        'GET_MESSAGE': {'method': 'get', 'url': '/api/v3/fleet/messages/get/'},
        'ADD_MESSAGE': {'method': 'post', 'url': '/api/v3/fleet/messages/add/'},
        'ARCHIVE_MESSAGE': {'method': 'post', 'url': '/api/v3/fleet/messages/archive/'},
        'DELETE_MESSAGE': {'method': 'post', 'url': '/api/v3/fleet/messages/delete/'},
        'PREPARE_TUNNEL': {'method': 'post', 'url': '/api/v3/fleet/proxy/prepare-tunnel/'},
        'SET_PROFILES': {'method': 'post', 'url': '/api/v3/fleet/profiles/set/'},
        'CHECK_TOKEN': {'method': 'post', 'url': '/api/v3/users/check-token/'},
        'GET_RELEASE': {'method': 'get', 'url': '/api/v3/packaging/check-for-update/'}
    }
}



'''
Miris Manager long polling management
This module is not intended to be used directly, only the client class should be used.
'''
import datetime
import logging
import os
import signal
import sys
import time
import traceback

from .signing import check_signature

logger = logging.getLogger('mm_client.lib.long_polling')


class LongPollingManager():

    def __init__(self, client):
        self.client = client
        self.run_systemd_notify = False
        self.last_error = None
        self.loop_running = False

    def loop(self, single_loop=False):
        # Check if systemd-notify should be called
        self.run_systemd_notify = self.client.conf.get('WATCHDOG') and os.system('which systemd-notify') == 0
        # Start connection loop
        logger.info('Starting long polling to %s' % self.client.conf['SERVER_URL'])
        self.loop_running = True

        def exit_handler(*args, **kwargs):
            self.loop_running = False
            logger.info('Long polling loop stopped')
            sys.exit(1)

        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGTERM, exit_handler)

        while self.loop_running:
            start = datetime.datetime.utcnow()
            success = self.call_long_polling()
            if single_loop:
                break
            if not success:
                # Avoid starting too often new connections
                duration = (datetime.datetime.utcnow() - start).seconds
                if duration < 5:
                    time.sleep(5 - duration)

    def call_long_polling(self):
        success = False
        try:
            logger.debug('Make long polling request')
            response = self.client.api_request('LONG_POLLING', timeout=300)
        except Exception as e:
            if 'timeout=300' not in str(e):
                msg = 'Long polling connection failed: %s: %s' % (e.__class__.__name__, e)
                if self.last_error == e.__class__.__name__:
                    logger.debug(msg)  # Avoid spamming
                else:
                    logger.error(msg)
                    self.last_error = e.__class__.__name__
        else:
            self.last_error = None
            if response:
                logger.info('Received long polling response: %s', response)
                success = True
                uid = response.get('uid')
                try:
                    status, data = self.process_long_polling(response)
                except Exception as e:
                    success = False
                    logger.error('Failed to process response: %s\n%s', e, traceback.format_exc())
                    self.client.set_command_status(uid, 'FAILED', str(e))
                    if os.environ.get('CI_PIPELINE_ID'):
                        # Propagate exception so that it can be detected in CI
                        raise
                else:
                    self.client.set_command_status(uid, status, data)
        finally:
            if self.run_systemd_notify:
                logger.debug('Notifying systemd watchdog.')
                os.system('systemd-notify WATCHDOG=1')
        return success

    def process_long_polling(self, response):
        logger.debug('Processing response.')
        if self.client.conf.get('API_KEY'):
            invalid = check_signature(self.client.conf, response)
            if invalid:
                raise ValueError('Invalid signature: %s' % invalid)
        uid = response.get('uid')
        action = response.get('action')
        if not action:
            raise ValueError('No action received.')
        params = response.get('params', {})
        logger.debug('Received command "%s": %s.', uid, action)
        if action == 'PING':
            return 'DONE', ''
        status, data = self.client.handle_action(uid=uid, action=action, params=params)
        if status not in ('DONE', 'IN_PROGRESS', 'FAILED'):
            logger.error('Your client has returned an invalid status in "handle_action".')
            raise ValueError('An error occurred during the processing of the action by the client.')
        if data is not None and not isinstance(data, str):
            logger.error('Your client has returned an invalid type for data in "handle_action".')
            raise ValueError('An error occurred during the processing of the action by the client.')
        return status, data



'''
Miris Manager client main module
'''
import logging
from pathlib import Path

import requests

from .lib import configuration as configuration_lib
from .lib import info as info_lib
from .lib import long_polling as long_polling_lib
from .lib import signing as signing_lib
from .lib import ssh_tunnel as ssh_tunnel_lib

logger = logging.getLogger('mm_client.client')


class MirisManagerRequestError(Exception):
    def __init__(self, *args, **kwargs):
        self.status_code = kwargs.pop('status_code', None)
        self.error_code = kwargs.pop('error_code', None)
        super().__init__(*args, **kwargs)


class MirisManagerClient():
    '''
    Miris Manager client class
    '''
    DEFAULT_CONF = None  # can be either a dict or a path (`str` object)

    def __init__(self, local_conf=None, setup_logging=True):
        # "local_conf" can be either a dict or a path (`str` object)
        # Setup logging
        if setup_logging:
            log_format = '%(asctime)s %(name)s %(levelname)s %(message)s'
            logging.basicConfig(level=logging.INFO, format=log_format)
        # Read conf file
        self.conf_checked = False
        self.conf = self.load_conf(local_conf)
        # Configure logging
        if setup_logging:
            level = getattr(logging, self.conf['LOG_LEVEL']) if self.conf.get('LOG_LEVEL') else logging.INFO
            root_logger = logging.getLogger('root')
            root_logger.setLevel(level)
            logger.setLevel(level)
            logging.captureWarnings(False)
            logger.debug('Logging conf set.')
        if not self.conf['VERIFY_SSL']:
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self._long_polling_manager = None
        self._ssh_tunnel_manager = None

    def load_conf(self, local_conf):
        self.local_conf = local_conf
        conf = configuration_lib.load_conf(self.DEFAULT_CONF, self.local_conf)
        self.conf_checked = False
        return conf

    def update_conf(self, key, value):
        self.conf[key] = value
        # write change in local_conf if it is a path
        configuration_lib.update_conf(self.local_conf, key, value)

    def check_conf(self):
        if not self.conf_checked:
            configuration_lib.check_conf(self.conf)
            self.conf_checked = True

    def get_url_info(self, url_or_action):
        if url_or_action.startswith('/'):
            return {'url': url_or_action}
        if url_or_action not in self.conf['API_CALLS']:
            raise MirisManagerRequestError(
                f'Invalid url requested: {url_or_action} does not exist in API_CALLS configuration.',
                status_code=0,
                error_code='invalid_url'
            )
        return self.conf['API_CALLS'][url_or_action]

    def _request(self, url, method='get', headers=None, params=None,
                 data=None, files=None, anonymous=None, timeout=None):
        req = getattr(requests, method)(
            url=self.conf['SERVER_URL'] + url,
            headers=headers,
            params=params,
            data=data,
            files=files,
            proxies=self.conf.get('PROXIES'),
            verify=self.conf['VERIFY_SSL'],
            timeout=timeout or self.conf['TIMEOUT']
        )
        status_code = req.status_code
        error_code = None
        body = req.text.strip()
        if req.status_code != 200:
            try:
                response = req.json()
                error = response['error']
                error_code = response.get('code')
            except Exception:
                error = 'Request failed with status code %s:\n%s.' % (req.status_code, body[:200])
            raise MirisManagerRequestError(
                error,
                status_code=status_code,
                error_code=error_code
            )
        response = req.json() if body else {}
        return response

    def _register(self):
        if self.conf.get('API_KEY'):
            return
        logger.info('No API key in configuration, requesting system registration...')
        data = info_lib.get_host_info(self.conf['SERVER_URL'])
        data['capabilities'] = ' '.join(self.conf['CAPABILITIES'])
        # Make API request
        response = self._request(self.get_url_info('REGISTER_SYSTEM')['url'], method='post', data=data)
        # Check response
        secret_key = response.get('secret_key')
        if not secret_key:
            raise MirisManagerRequestError(
                'No secret key received.',
                status_code=200,
                error_code='no_secret'
            )
        api_key = response.get('api_key')
        if not api_key:
            raise MirisManagerRequestError(
                'No API key received.',
                status_code=200,
                error_code='no_api_key'
            )
        self.update_conf('SECRET_KEY', secret_key)
        self.update_conf('API_KEY', api_key)
        logger.info('System registration done.')
        return True

    def api_request(self, url_or_action, method='get', headers=None, params=None,
                    data=None, files=None, anonymous=None, timeout=None):
        self.check_conf()
        url_info = self.get_url_info(url_or_action)
        if anonymous is None:
            anonymous = bool(url_info.get('anonymous'))
        if anonymous:
            _headers = headers
        else:
            # Register system if no API key and auto registration
            if not self.conf.get('API_KEY'):
                if not self.conf['AUTO_REGISTRATION']:
                    raise ValueError('The client auto registration is disabled and no API_KEY is set in conf file, '
                                     'please set one or turn on auto registration.')
                try:
                    self._register()
                except Exception as e:
                    logger.error('Registration failed: %s', e)
                    raise
            # Add signature in headers
            # headers with "_" are ignored by Django
            _headers = {'api-key': self.conf['API_KEY']}
            if not anonymous:
                signature = signing_lib.get_signature(self.conf)
                if signature:
                    _headers.update(signature)
            if headers:
                _headers.update(headers)
        # Make API request
        response = self._request(
            url_info['url'],
            method=url_info.get('method', method),
            headers=_headers,
            params=params,
            data=data,
            files=files,
            timeout=timeout
        )
        return response

    def long_polling_loop(self, single_loop=False):
        if not self._long_polling_manager:
            self._long_polling_manager = long_polling_lib.LongPollingManager(self)
        self._long_polling_manager.loop(single_loop)

    def handle_action(self, uid, action, params):
        '''
        Function that should be implemented in your client to process the long polling responses.
        IMPORTANT: Any code written here should not be blocking more than 5s because of the
                   delay after which the system is considered as offline in Miris Manager.
        Arguments:
        - uid: The system command unique identifier.
        - action: The action to run.
        - params: The action parameters.
        Must return a tuple: (status, data)
        - status: The system command status (string). Possible values:
        - "DONE": The command has been executed successfully.
        - "IN_PROGRESS": The command has been started but is not yet completed.
        - "FAILED": The command execution has failed.
        - data: The command result data (string). It can be a json dump or a message. Empty strings are allowed.
        '''
        raise NotImplementedError('Your class should override the "handle_action" method.')

    def set_command_status(self, command_uid, status='DONE', data=None):
        if not command_uid:
            return
        try:
            self.api_request('SET_COMMAND_STATUS', data=dict(
                uid=command_uid,
                status=status,
                data=data or '',
            ))
        except Exception as e:
            logger.error('Unable to communicate command status: %s %s', type(e), e)

    def set_info(self):
        data = info_lib.get_host_info(self.conf['SERVER_URL'])
        data['capabilities'] = ' '.join(self.conf['CAPABILITIES'])
        # Make API request
        response = self.api_request('SET_INFO', data=data)
        return response

    def update_capabilities(self):
        data = {
            'capabilities': ' '.join(self.conf['CAPABILITIES']),
        }
        # Make API request
        response = self.api_request('SET_INFO', data=data)
        return response

    def set_status(self, status=None, status_info=None, status_message=None,
                   profile=None, remaining_space=None, remaining_time=None):
        data = {}
        if status is not None:
            data['status'] = status
        if status_info is not None:
            data['status_info'] = status_info
        if status_message is not None or status is not None:
            data['status_message'] = status_message or ''
        if profile is not None:
            data['profile'] = profile
        if remaining_space == 'auto':
            remaining_space = info_lib.get_remaining_space()
        if remaining_space is not None:
            data['remaining_space'] = remaining_space
        if remaining_time is not None:
            data['remaining_time'] = remaining_time
        if not data:
            raise ValueError('No data to update.')
        response = self.api_request('SET_STATUS', data=data)
        return response

    def set_screenshot(self, path, file_name=None):
        with open(path, 'rb') as file_obj:
            response = self.api_request('SET_SCREENSHOT', files=dict(
                screenshot=(file_name or Path(path).name, file_obj)
            ))
        return response

    def open_tunnel(self, status_callback=None):
        if not self._ssh_tunnel_manager:
            self._ssh_tunnel_manager = ssh_tunnel_lib.SSHTunnelManager(self, status_callback)
        self._ssh_tunnel_manager.tunnel_loop()

    def close_tunnel(self):
        if self._ssh_tunnel_manager:
            self._ssh_tunnel_manager.close_tunnel()



import json
from unittest.mock import patch

CONFIG = {
    'SERVER_URL': 'https://mmctest',
    'SECRET_KEY': 'the secret key',
    'API_KEY': 'test API key',
}


class MockResponse:
    def __init__(self, json_data, status_code):
        self.text = json.dumps(json_data)
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


def mocked_request(*args, **kwargs):
    url = kwargs['url']
    if url == CONFIG['SERVER_URL'] + '/api/':
        return MockResponse({'version': '8.0.0'}, 200)
    if url == CONFIG['SERVER_URL'] + '/api/v3/fleet/control/set-command-status/':
        return MockResponse({}, 200)
    if url == CONFIG['SERVER_URL'] + '/remote-event/v3':
        from mm_client.lib.signing import get_signature
        data = get_signature(CONFIG)
        data.update({
            'uid': 'test_uid',
            'action': 'START_RECORDING',
            'params': {'channel': 'Chan'},
        })
        return MockResponse(data, 200)
    print(f'Non mocked URL: {url}')
    return MockResponse(None, 404)


@patch('requests.post', side_effect=mocked_request)
@patch('requests.get', side_effect=mocked_request)
def test_client(mock_get, mock_post):
    from mm_client.client import MirisManagerClient
    mmc = MirisManagerClient(local_conf=CONFIG)
    response = mmc.api_request('PING')
    assert isinstance(response, dict)
    assert response['version'] == '8.0.0'

    assert len(mock_get.call_args_list) == 1
    assert len(mock_post.call_args_list) == 0


@patch('requests.post', side_effect=mocked_request)
@patch('requests.get', side_effect=mocked_request)
def test_long_polling(mock_get, mock_post):
    from mm_client.client import MirisManagerClient

    commands = []

    class LongPollingClient(MirisManagerClient):
        DEFAULT_CONF = CONFIG

        def handle_action(self, uid, action, params):
            commands.append((uid, action, params))
            return 'DONE', ''

    mmc = LongPollingClient(local_conf=CONFIG)
    mmc.long_polling_loop(single_loop=True)

    assert len(mock_get.call_args_list) == 1
    assert len(mock_post.call_args_list) == 1
    call = mock_post.call_args_list[0]
    assert call.kwargs['data'] == {'uid': 'test_uid', 'status': 'DONE', 'data': ''}

    assert commands == [('test_uid', 'START_RECORDING', {'channel': 'Chan'})]





def test_get_host_info():
    from mm_client.lib.info import get_host_info

    expected = ['hostname', 'local_ip', 'mac']

    info = get_host_info(url='https://localhost')
    assert sorted(info.keys()) == expected
    for field in expected:
        assert info[field]


def test_get_free_space_bytes():
    from mm_client.lib.info import get_free_space_bytes

    remaining = get_free_space_bytes('/home')
    assert remaining > 0


def test_get_remaining_space():
    from mm_client.lib.info import get_remaining_space

    remaining = get_remaining_space()
    assert remaining > 0



'''
Miris Manager client signing functions
This module is not intended to be used directly, only the client class should be used.
'''
import base64
import datetime
import hashlib
import hmac
import logging

logger = logging.getLogger('mm_client.lib.signing')


def get_signature(conf):
    if not conf.get('SECRET_KEY') or not conf.get('API_KEY'):
        return {}
    utime = datetime.datetime.utcnow().strftime('%Y-%m-%d_%H-%M-%S_%f')
    to_sign = 'time=%s|api_key=%s' % (utime, conf['API_KEY'])
    hm = hmac.new(
        conf['SECRET_KEY'].encode('utf-8'),
        msg=to_sign.encode('utf-8'),
        digestmod=hashlib.sha256
    ).digest()
    hm = base64.b64encode(hm).decode('utf-8')
    return {'time': utime, 'hmac': hm}


def check_signature(conf, rdata):
    if not conf.get('SECRET_KEY') or not conf.get('API_KEY'):
        return None
    remote_time = rdata.get('time')
    remote_hmac = rdata.get('hmac')
    if not remote_time or not remote_hmac:
        return 'some mandatory data are missing.'
    try:
        rdate = datetime.datetime.strptime(remote_time, '%Y-%m-%d_%H-%M-%S_%f')
    except ValueError:
        return 'the received time is invalid.'
    try:
        rhmac = base64.b64decode(remote_hmac)
    except Exception:
        return 'the received hmac is invalid.'
    utcnow = datetime.datetime.utcnow()
    diff = utcnow - rdate if utcnow > rdate else rdate - utcnow
    if diff.seconds > 300:
        return 'the difference between the request time and the current time is too large.'
    to_sign = 'time=%s|api_key=%s' % (remote_time, conf['API_KEY'])
    hm = hmac.new(
        conf['SECRET_KEY'].encode('utf-8'),
        msg=to_sign.encode('utf-8'),
        digestmod=hashlib.sha256
    ).digest()
    if rhmac != hm:
        return 'the received and computed HMAC values do not match.'
    return None



'''
Miris Manager SSH tunnel management
This module is not intended to be used directly, only the client class should be used.

The SSH tunnel goal is to access the system web interface (HTTPS) from
Miris Manager using a connection from the system to the Miris Manager.
'''
import logging
import os
import subprocess
import time
import multiprocessing
import re
import signal
from pathlib import Path

logger = logging.getLogger('mm_client.lib.ssh_tunnel')


class MirisManagerTunnelError(Exception):
    pass


def get_ssh_public_key():
    ssh_dir = Path('~/.ssh').expanduser()
    ssh_key_path = ssh_dir / 'miris-manager-client-key'
    ssh_pub_path = ssh_dir / 'miris-manager-client-key.pub'
    if not ssh_dir.exists():
        ssh_dir.mkdir(parents=True)
        ssh_dir.chmod(0o700)
    if ssh_key_path.exists():
        if not ssh_pub_path.exists():
            raise MirisManagerTunnelError(
                f'Weird state detetected: "{ssh_key_path}" exists but not "{ssh_pub_path}" !'
            )
        logger.debug('Using existing SSH key: "%s".', ssh_key_path)
    else:
        logger.info('Creating new SSH key: "%s".', ssh_key_path)
        p = subprocess.Popen(
            ['ssh-keygen', '-t', 'rsa', '-b', '4096', '-f', str(ssh_key_path), '-N', ''],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )
        p.communicate(input=b'\n\n\n')
        if p.returncode != 0:
            out = p.stdout.decode('utf-8').strip()
            raise MirisManagerTunnelError(
                f'Failed to generate SSH key:\n{out}'
            )
        ssh_key_path.chmod(0o600)
        ssh_pub_path.chmod(0o600)
    public_key = ssh_pub_path.read_text()
    return public_key


def prepare_ssh_command(host, info):
    ssh_key_path = Path('~/.ssh/miris-manager-client-key').expanduser()
    command = ['ssh',
               '-i', str(ssh_key_path),
               '-nvNT',
               '-o', 'IdentitiesOnly=yes',
               '-o', 'NumberOfPasswordPrompts=0',
               '-o', 'CheckHostIP=no',
               '-o', 'StrictHostKeyChecking=no',
               '-o', 'ServerAliveInterval=10',
               '-R', '%s:127.0.0.1:443' % info['control_port'],
               '-R', '%s:127.0.0.1:22' % info['maintenance_port'],
               '-p', str(info['ssh_port']),
               '%s@%s' % (info['ssh_user'], host)]
    return command


class SSHTunnelManager():

    def __init__(self, client, status_callback=None):
        self.client = client
        self.status_callback = status_callback
        self.pattern_list = [
            dict(id='connecting', pattern=re.compile(
                r'debug1: Connecting to (?P<hostname>[^ ]+) \[(?P<ip>[0-9\.]{7,15})\] port (?P<port>\d{1,5}).\r\n'
            )),
            dict(id='connected', pattern=re.compile(
                r'debug1: Connection established.\r\n'
            )),
            dict(id='authenticated', pattern=re.compile(
                r'debug1: Authentication succeeded \((?P<method>[^\)]+)\).\r\n'
            )),
            dict(id='authenticated', pattern=re.compile(
                r'Authenticated to (?P<hostname>[^ ]+) \(\[(?P<ip>[0-9\.]{7,15})\]:(?P<port>\d{1,5})\).\r\n'
            )),
            dict(id='running', pattern=re.compile(
                r'debug1: Entering interactive session.\r\n'
            )),
            dict(id='not_known', pattern=re.compile(
                r'ssh: [^:]+: Name or service not known\r\n'
            )),
            dict(id='port_refused', pattern=re.compile(
                r'Warning: remote port forwarding failed for listen port (?P<port>\d{1,5})'
            )),
            dict(id='refused', pattern=re.compile(
                r'ssh: connect to host [^:]+: Connection refused\r\n'
            )),
            dict(id='control_refused', pattern=re.compile(
                r'connect_to (?P<hostname>[^ ]+) port (?P<port>\d{1,5}): failed\.\r\n'
            )),
            dict(id='denied', pattern=re.compile(
                r'Permission denied \(publickey,password\).\r\n'
            )),
            dict(id='closed', pattern=re.compile(
                r'Connection to (?P<hostname>[^ ]+) closed.\r\n'
            )),
        ]
        self.loop_ssh_tunnel = False
        self.process = None
        self.stdout_queue = None
        self.stdout_reader = None
        self.stderr_reader = None
        self.ssh_tunnel_state = {
            'ssh_user': 'skyreach',
            'ssh_port': 22,
            'control_port': 0,
            'maintenance_port': 0,
            'state': 'Not running',
            'command': '',
            'last_tunnel_info': ''
        }

    def establish_tunnel(self):
        public_key = None
        response = None
        logger.debug('Establishing new tunnel to %s', self.client.conf['SERVER_URL'])
        self._stop_reader()
        self._try_closing_process()
        self.update_ssh_state('state', 'prepare tunnel')
        try:
            logger.debug('Prepare tunnel')
            public_key = get_ssh_public_key()
            response = self.client.api_request('PREPARE_TUNNEL', data=dict(public_key=public_key))
        except Exception as e:
            self.update_ssh_state('state', 'prepare tunnel failed')
            self.update_ssh_state('control_port', 0)
            self.update_ssh_state('maintenance_port', 0)
            self.update_ssh_state('command', ['PREPARE_TUNNEL', self.client.conf['SERVER_URL']])
            logger.error('Cannot prepare ssh tunnel : %s', str(e))
            return
        ssh_user = response.get('ssh_user')
        if ssh_user and ssh_user != self.ssh_tunnel_state['ssh_user']:
            self.update_ssh_state('ssh_user', response['ssh_user'])
        ssh_port = response.get('ssh_port')
        if ssh_port and ssh_port != self.ssh_tunnel_state['ssh_port']:
            self.update_ssh_state('ssh_port', response['ssh_port'])
        maintenance_port = response.get('maintenance_port')
        if maintenance_port and maintenance_port != self.ssh_tunnel_state['maintenance_port']:
            self.update_ssh_state('maintenance_port', response['maintenance_port'])
        port = response.get('control_port') or response.get('port')
        if port is not None:
            self.update_ssh_state('control_port', port)
            host = self.client.conf['SERVER_URL'].split('://', 1)[-1]
            host = host.rsplit(':', 1)[0].rstrip('/')
            cmd = prepare_ssh_command(host, self.ssh_tunnel_state)
            self.update_ssh_state('command', cmd)
            logger.info('Starting SSH with command:\n    %s', ' '.join(cmd))
            if self.loop_ssh_tunnel:
                self.process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    preexec_fn=os.setsid
                )
                self.stdout_queue = multiprocessing.Queue()
                self.stdout_reader = AsynchronousFileReader(self.process.stdout, self.stdout_queue)
                self.stdout_reader.start()
                self.stderr_reader = AsynchronousFileReader(self.process.stderr, self.stdout_queue)
                self.stderr_reader.start()
                return True
        else:
            logger.debug('No control port provided, not starting ssh tunnel')

    def update_ssh_state(self, key, value):
        if key == 'state' and self.ssh_tunnel_state.get('state') != value:
            logger.info('SSH state changed to %s', value)
        if self.ssh_tunnel_state.get(key) is not None:
            self.ssh_tunnel_state[key] = value
        else:
            logger.warning('Key %s not exists in ssh state dict', key)
        if self.status_callback:
            self.status_callback(self.ssh_tunnel_state)

    def close_tunnel(self, thread_event=None):
        logger.debug('Close ssh tunnel asked')
        self.loop_ssh_tunnel = False
        if thread_event:
            thread_event.set()
        self._stop_reader()
        self._try_closing_process()

    def _try_closing_process(self):
        if self.process:
            timeout = 5
            logger.debug('Waiting %ss for ssh process to terminate' % timeout)
            self.process.terminate()
            while self.process and self.process.poll() is None and timeout != 0:
                timeout -= 1
                time.sleep(1)
            if timeout == 0:
                logger.warning('SSH tunnel has not terminated, trying to kill it')
                pgrp = os.getpgid(self.process.pid)
                os.killpg(pgrp, signal.SIGINT)
                # self.process.kill()
                logger.warning('SSH tunnel killed')
            else:
                logger.info('SSH tunnel subprocess terminated')
            self.process = None

    def _stop_reader(self):
        if self.stdout_reader is not None and self.stdout_reader.pid is not None:
            logger.debug('Waiting for stdout_reader subprocess to terminate')
            try:
                os.kill(self.stdout_reader.pid, signal.SIGKILL)
            except ProcessLookupError:
                pass
            self.stdout_reader = None
            logger.debug('SSH stdout_reader killed')

        if self.stderr_reader is not None and self.stderr_reader.pid is not None:
            logger.debug('Wait for stderr_reader subprocess to terminate')
            try:
                os.kill(self.stderr_reader.pid, signal.SIGKILL)
            except ProcessLookupError:
                pass
            self.stderr_reader = None
            logger.debug('SSH stderr_reader killed')

        if self.stdout_queue:
            logger.debug('Waiting for ssh queue to close')
            self.stdout_queue.close()
            self.stdout_queue.cancel_join_thread()
            self.stdout_queue = None
            logger.debug('SSH queue closed')

    def tunnel_loop(self, thread_event=None):
        check_delay = 1
        self.loop_ssh_tunnel = True
        self.update_ssh_state('state', 'loading')
        self.update_ssh_state('control_port', 0)
        self.update_ssh_state('maintenance_port', 0)
        self.update_ssh_state('command', ['Load', self.client.conf['SERVER_URL']])
        while self.loop_ssh_tunnel:
            need_retry = False
            if self.process is not None:
                need_retry = self.read_ssh_stdout()
            else:
                need_retry = True
            if need_retry:
                try:
                    success = self.establish_tunnel()
                except Exception as e:
                    logger.error('error while establishing tunnel %s', e)
                if need_retry or not success:
                    self._wait(thread_event)
            else:
                self._wait(thread_event, check_delay)

    def _wait(self, thread_event=None, delay=10):
        if thread_event:
            # blocking unless thread_event.set() is called
            thread_event.wait(delay)
        else:
            # use a normal sleep if not running in a thread
            time.sleep(delay)

    def read_ssh_stdout(self):
        need_retry = False
        return_code = self.process.poll()
        if return_code is not None:
            logger.debug('SSH process has terminated')
            ssh_logs = ''
            try:
                while not self.stdout_queue.empty():
                    ssh_logs += self.stdout_queue.get_nowait()
            except OSError as e:
                ssh_logs = str(e)
            self.update_ssh_state('state', 'error')
            self.update_ssh_state('last_tunnel_info', ssh_logs)
            logger.error('SSH tunnel process has terminated with: %s', ssh_logs)
            need_retry = True
        else:
            # process is still alive
            try:
                while not self.stdout_queue.empty():
                    ssh_stdout = self.stdout_queue.get_nowait()
                    pattern_id_found = None
                    for pattern_dict in self.pattern_list:
                        if pattern_dict['pattern'].match(ssh_stdout):
                            pattern_id_found = pattern_dict['id']
                            self.update_ssh_state('state', pattern_id_found)
                            self.update_ssh_state('last_tunnel_info', ssh_stdout)
                            break
                    if not pattern_id_found:
                        if ssh_stdout.startswith('debug1:') or ssh_stdout.startswith('OpenSSH_'):
                            logger.debug('[SSH stdout] %s', ssh_stdout)
                        else:
                            logger.warning('[SSH stdout] %s', ssh_stdout)
                    elif pattern_id_found not in ['connecting', 'connected', 'authenticated', 'running']:
                        logger.error(
                            (
                                'Need to retry tunnel '
                                '(ssh port: %s, remote control port: %s, remote maintenance port: %s) '
                                'because ssh command failed in stdout %s'
                            ),
                            self.ssh_tunnel_state.get('ssh_port'),
                            self.ssh_tunnel_state.get('control_port'),
                            self.ssh_tunnel_state.get('maintenance_port'),
                            pattern_id_found
                        )
                        need_retry = True
                        break
            except OSError as e:
                logger.error(e)
                need_retry = True
        return need_retry


class AsynchronousFileReader(multiprocessing.Process):
    def __init__(self, fd, data_queue):
        multiprocessing.Process.__init__(self)
        self._fd = fd
        self._queue = data_queue

    def run(self):
        while self.is_alive():
            line = self._fd.readline().decode('utf-8')
            if line:
                self._queue.put(line)
                continue
            time.sleep(.5)

    def eof(self):
        return not self.is_alive() and self._queue.empty()



from pathlib import Path

import pytest


@pytest.fixture()
def conf_path():
    path = Path('/tmp/mm-conf.json')
    path.write_text('{"SERVER_URL": "https://test"}')
    yield path
    path.unlink(missing_ok=True)


def test_conf_file__valid(conf_path):
    from mm_client.lib.configuration import load_conf, update_conf

    conf = load_conf(default_conf=conf_path)
    assert conf['SERVER_URL'] == 'https://test'

    updated = update_conf(conf_path, 'test', 'val')
    assert updated is True


def test_conf_file__does_not_exist(conf_path):
    from mm_client.lib.configuration import load_conf, update_conf

    conf_path.unlink()
    conf = load_conf(default_conf=conf_path)
    assert conf['SERVER_URL'] == 'https://mirismanager'

    updated = update_conf(conf_path, 'test', 'val')
    assert updated is True


def test_conf_dict():
    from mm_client.lib.configuration import load_conf, update_conf

    conf = load_conf(
        default_conf={'SERVER_URL': 'https://nope'},
        local_conf={'SERVER_URL': 'https://test'}
    )
    assert conf['SERVER_URL'] == 'https://test'

    updated = update_conf({'SERVER_URL': 'https://test'}, 'test', 'val')
    assert updated is False


def test_conf_default():
    from mm_client.lib.configuration import load_conf, update_conf

    conf = load_conf()
    assert conf['SERVER_URL'] == 'https://mirismanager'

    updated = update_conf(None, 'test', 'val')
    assert updated is False


@pytest.mark.parametrize('conf, is_valid', [
    pytest.param(
        {'SERVER_URL': 'https://mirismanager'},
        False,
        id='default'),
    pytest.param(
        {'SERVER_URL': 'https://test/'},
        True,
        id='valid'),
])
def test_conf_check(conf, is_valid):
    from mm_client.lib.configuration import check_conf

    if is_valid:
        check_conf(conf)
    else:
        with pytest.raises(ValueError):
            check_conf(conf)
